@@ -1,14 +1,13 @@
 import json
 import logging
 from ssl import SSLContext
-from typing import Any, Union
+from typing import Any, Union, List
 from typing import Dict, Optional
 from urllib.parse import quote
 
 import aiohttp
 from aiohttp import BasicAuth, ClientSession
 
-from slack_sdk.errors import SlackApiError
 from .internal_utils import (
     _build_request_headers,
     _debug_log_response,
@@ -35,6 +34,12 @@ from .user import User
 from .group import Group
 from ...proxy_env_variable_loader import load_http_proxy_from_env
 
+from slack_sdk.http_retry.async_handler import AsyncRetryHandler
+from slack_sdk.http_retry.builtin_async_handlers import async_default_handlers
+from slack_sdk.http_retry.request import HttpRequest as RetryHttpRequest
+from slack_sdk.http_retry.response import HttpResponse as RetryHttpResponse
+from slack_sdk.http_retry.state import RetryState
+
 
 class AsyncSCIMClient:
     BASE_URL = "https://api.slack.com/scim/v1/"
@@ -49,6 +54,7 @@ class AsyncSCIMClient:
     auth: Optional[BasicAuth]
     default_headers: Dict[str, str]
     logger: logging.Logger
+    retry_handlers: List[AsyncRetryHandler]
 
     def __init__(
         self,
@@ -64,21 +70,25 @@ class AsyncSCIMClient:
         user_agent_prefix: Optional[str] = None,
         user_agent_suffix: Optional[str] = None,
         logger: Optional[logging.Logger] = None,
+        retry_handlers: Optional[List[AsyncRetryHandler]] = None,
     ):
         """API client for SCIM API
         See https://api.slack.com/scim for more details
-        :param token: An admin user's token, which starts with xoxp-
-        :param timeout: request timeout (in seconds)
-        :param ssl: ssl.SSLContext to use for requests
-        :param proxy: proxy URL (e.g., localhost:9000, http://localhost:9000)
-        :param base_url: the base URL for API calls
-        :param session: a complete aiohttp.ClientSession
-        :param trust_env_in_session: True/False for aiohttp.ClientSession
-        :param auth: Basic auth info for aiohttp.ClientSession
-        :param default_headers: request headers to add to all requests
-        :param user_agent_prefix: prefix for User-Agent header value
-        :param user_agent_suffix: suffix for User-Agent header value
-        :param logger: custom logger
+
+        Args:
+            token: An admin user's token, which starts with `xoxp-`
+            timeout: Request timeout (in seconds)
+            ssl: `ssl.SSLContext` to use for requests
+            proxy: Proxy URL (e.g., `localhost:9000`, `http://localhost:9000`)
+            base_url: The base URL for API calls
+            session: `aiohttp.ClientSession` instance
+            trust_env_in_session: True/False for `aiohttp.ClientSession`
+            auth: Basic auth info for `aiohttp.ClientSession`
+            default_headers: Request headers to add to all requests
+            user_agent_prefix: Prefix for User-Agent header value
+            user_agent_suffix: Suffix for User-Agent header value
+            logger: Custom logger
+            retry_handlers: Retry handlers
         """
         self.token = token
         self.timeout = timeout
@@ -93,6 +103,9 @@ class AsyncSCIMClient:
             user_agent_prefix, user_agent_suffix
         )
         self.logger = logger if logger is not None else logging.getLogger(__name__)
+        self.retry_handlers = (
+            retry_handlers if retry_handlers is not None else async_default_handlers()
+        )
 
         if self.proxy is None or len(self.proxy.strip()) == 0:
             env_variable = load_http_proxy_from_env(self.logger)
@@ -291,14 +304,6 @@ class AsyncSCIMClient:
             body_params = json.dumps(body_params)
         headers["Content-Type"] = "application/json;charset=utf-8"
 
-        if self.logger.level <= logging.DEBUG:
-            headers_for_logging = {
-                k: "(redacted)" if k.lower() == "authorization" else v
-                for k, v in headers.items()
-            }
-            self.logger.debug(
-                f"Sending a request - url: {url}, params: {body_params}, headers: {headers_for_logging}"
-            )
         session: Optional[ClientSession] = None
         use_running_session = self.session and not self.session.closed
         if use_running_session:
@@ -310,7 +315,8 @@ class AsyncSCIMClient:
                 trust_env=self.trust_env_in_session,
             )
 
-        resp: SCIMResponse
+        last_error: Optional[Exception] = None
+        resp: Optional[SCIMResponse] = None
         try:
             request_kwargs = {
                 "headers": headers,
@@ -318,25 +324,105 @@ class AsyncSCIMClient:
                 "ssl": self.ssl,
                 "proxy": self.proxy,
             }
-            async with session.request(http_verb, url, **request_kwargs) as res:
-                response_body = {}
-                try:
-                    response_body = await res.text()
-                except aiohttp.ContentTypeError:
+            retry_request = RetryHttpRequest(
+                method=http_verb,
+                url=url,
+                headers=headers,
+                body_params=body_params,
+            )
+
+            retry_state = RetryState()
+            counter_for_safety = 0
+            while counter_for_safety < 100:
+                counter_for_safety += 1
+                # If this is a retry, the next try started here. We can reset the flag.
+                retry_state.next_attempt_requested = False
+                retry_response: Optional[RetryHttpResponse] = None
+                response_body = ""
+
+                if self.logger.level <= logging.DEBUG:
+                    headers_for_logging = {
+                        k: "(redacted)" if k.lower() == "authorization" else v
+                        for k, v in headers.items()
+                    }
                     self.logger.debug(
-                        f"No response data returned from the following API call: {url}."
+                        f"Sending a request - url: {url}, params: {body_params}, headers: {headers_for_logging}"
                     )
-                except json.decoder.JSONDecodeError as e:
-                    message = f"Failed to parse the response body: {str(e)}"
-                    raise SlackApiError(message, res)
+
+                try:
+                    async with session.request(http_verb, url, **request_kwargs) as res:
+                        try:
+                            response_body = await res.text()
+                            retry_response = RetryHttpResponse(
+                                status_code=res.status,
+                                headers=res.headers,
+                                data=response_body.encode("utf-8")
+                                if response_body is not None
+                                else None,
+                            )
+                        except aiohttp.ContentTypeError:
+                            self.logger.debug(
+                                f"No response data returned from the following API call: {url}."
+                            )
+
+                        if res.status == 429:
+                            for handler in self.retry_handlers:
+                                if await handler.can_retry_async(
+                                    state=retry_state,
+                                    request=retry_request,
+                                    response=retry_response,
+                                ):
+                                    if self.logger.level <= logging.DEBUG:
+                                        self.logger.info(
+                                            f"A retry handler found: {type(handler).__name__} "
+                                            f"for {http_verb} {url} - rate_limited"
+                                        )
+                                    await handler.prepare_for_next_attempt_async(
+                                        state=retry_state,
+                                        request=retry_request,
+                                        response=retry_response,
+                                    )
+                                    break
+
+                        if retry_state.next_attempt_requested is False:
+                            resp = SCIMResponse(
+                                url=url,
+                                status_code=res.status,
+                                raw_body=response_body,
+                                headers=res.headers,
+                            )
+                            _debug_log_response(self.logger, resp)
+                            return resp
+
+                except Exception as e:
+                    last_error = e
+                    for handler in self.retry_handlers:
+                        if await handler.can_retry_async(
+                            state=retry_state,
+                            request=retry_request,
+                            response=retry_response,
+                            error=e,
+                        ):
+                            if self.logger.level <= logging.DEBUG:
+                                self.logger.info(
+                                    f"A retry handler found: {type(handler).__name__} "
+                                    f"for {http_verb} {url} - {e}"
+                                )
+                            await handler.prepare_for_next_attempt_async(
+                                state=retry_state,
+                                request=retry_request,
+                                response=retry_response,
+                                error=e,
+                            )
+                            break
+
+                    if retry_state.next_attempt_requested is False:
+                        raise last_error
+
+            if resp is not None:
+                return resp
+            raise last_error
 
-                resp = SCIMResponse(
-                    url=url,
-                    status_code=res.status,
-                    raw_body=response_body,
-                    headers=res.headers,
-                )
-                _debug_log_response(self.logger, resp)
         finally:
             if not use_running_session:
                 await session.close()
